/* CDLL: circular doubly-linked list.
   in TBCppA project: tracer-based CPP analyzer.
    Copyright (c) 2006 Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
    $Id: list.c,v 1.4 2006/11/01 11:36:57 gondow Exp gondow $ */

/* CDLL: circular doubly-linked list.

   The first CDLL element is always used as anchor, i.e., 
   a pointer to the list, and logically not a member of the list.
       null list: CDLL
       singleton: CDLL->CDLL
       3 elems. : CDLL->CDLL->CDLL->CDLL
                 anchor head        tail
    anchor's data is always ignored (not used for any purpose).
    This structure makes this file's code simpler. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "list.h"
#include "misc.h"
/* ================================================================== */
void
CDLL_dump  (struct CDLL *head)
{
    int i;
    struct CDLL *p;

    if (head == NULL) return;
    for (i = 0, p = head->next; p != head; i++, p = p->next) {
	fprintf (stderr, "%d: %s\n", i, p->param_name);
    }
}

void
CDLL_init   (struct CDLL *head)
{
    head->param_name = NULL;
    head->depth      = INT_MIN;
    head->type       = 0;
    head->next = head->prev = head;
}

void
CDLL_insert (struct CDLL *head, char *param_name)
{
    struct CDLL *cdll = emalloc (sizeof (struct CDLL));
    assert (cdll != NULL);
    cdll->param_name = param_name;
    cdll->prev = head->prev;
    cdll->next = head;
    head->prev->next = cdll;
    head->prev = cdll;
}

struct CDLL *
CDLL_append  (struct CDLL *list1, struct CDLL *list2)
{
    if (list2 == NULL || CDLL_is_null (list2))
	return list1;
    if (list1 == NULL || CDLL_is_null (list1))
	return list2;

    list1->prev->next = list2->next; /* modify list1's tail */
    list2->next->prev = list1->prev; /* modify list2's second */
    list1->prev       = list2->prev; /* modify list1's head */
    list2->prev->next = list1;       /* modify list2's tail */
#if 0
    EFREE (list2);
#endif

    return list1;
}

int
CDLL_query  (struct CDLL *head, char *param_name)
{
    int i;
    struct CDLL *p;
    assert (param_name != NULL);
    if (head == NULL)
	return -1;
    for (i = 0, p = head->next; p != head; i++, p = p->next) {
	if (!strcmp (param_name, p->param_name))
	    return i;
    }
    return -1; /* not found */
}

void
CDLL_free   (struct CDLL *head)
{
    struct CDLL *p, *next;
    assert (head != NULL);
    for (p = head->next; p != head; p = next) {
        next = p->next;
	p->next->prev = p->prev;
	p->prev->next = p->next;
	p->next = p->prev = NULL;
	EFREE (p->param_name);
	EFREE (p);
    }
    assert (head->next == head);
    assert (head->prev == head);
}

int
CDLL_is_null     (struct CDLL *head)
{
    return head == head->next;
}

#if 0
void
CDLL_unparse (struct CDLL *head, FILE *fp)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next) {
	if (p != head->next)
	    fputs (", ", fp);
	fputs (p->param_name, fp);
    }
}
#endif

struct CDLL *
CDLL_pop  (struct CDLL *head)
{
    struct CDLL *ret_cdll;

    if (head == NULL || CDLL_is_null (head))
	return NULL;

    ret_cdll = head->next;
    assert (ret_cdll != NULL);

    ret_cdll->prev->next = ret_cdll->next;
    ret_cdll->next->prev = ret_cdll->prev;

    return ret_cdll;
}

void
CDLL_top  (struct CDLL *head, int *depth, int *type)
{
    if (head == NULL || CDLL_is_null (head)) {
	*depth = INT_MIN;
	*type  = 0;
	return;
    }
    *depth = head->next->depth;
    *type  = head->next->type;
    return;
}

void
CDLL_push (struct CDLL *head, int depth, int type)
{
    struct CDLL *cdll = emalloc (sizeof (struct CDLL));
    assert (cdll != NULL);
    cdll->depth = depth;
    cdll->type = type;
    cdll->prev = head;
    cdll->next = head->next;
    head->next->prev = cdll;
    head->next = cdll;
}
/* ================================================================== */
#if 0
int
main (void)
{
    struct CDLL head, head2, *cdll;
#if 0
    CDLL_init (&head);
    CDLL_init (&head2);

    CDLL_insert (&head, estrdup ("param1"));
    CDLL_insert (&head, estrdup ("param2"));
    CDLL_insert (&head, estrdup ("param3"));
    CDLL_insert (&head, estrdup ("param4"));
    CDLL_dump (&head);
    printf ("query: %d\n", CDLL_query (&head, "param1"));
    printf ("query: %d\n", CDLL_query (&head, "param2"));
    printf ("query: %d\n", CDLL_query (&head, "param3"));
    printf ("query: %d\n", CDLL_query (&head, "param4"));
    printf ("query: %d\n", CDLL_query (&head, "param5"));
    // CDLL_unparse (&head, stderr);

    CDLL_insert (&head2, estrdup ("param21"));
    CDLL_insert (&head2, estrdup ("param22"));
    CDLL_insert (&head2, estrdup ("param23"));
    CDLL_insert (&head2, estrdup ("param24"));

    CDLL_append (&head, &head2);
    CDLL_dump (&head);
    CDLL_free (&head);
#endif

#if 0
    CDLL_init (&head);
    CDLL_push (&head, 10);
    CDLL_push (&head, 20);
    CDLL_push (&head, 30);
    CDLL_push (&head, 40);
    printf ("%d\n", CDLL_top (&head));
    printf ("%d\n", CDLL_pop (&head)->depth);
    printf ("%d\n", CDLL_top (&head));
    printf ("%d\n", CDLL_pop (&head)->depth);
    printf ("%d\n", CDLL_top (&head));
    printf ("%d\n", CDLL_pop (&head)->depth);
    printf ("%d\n", CDLL_top (&head));
    printf ("%d\n", CDLL_pop (&head)->depth);
    printf ("%d\n", CDLL_is_null (&head));
#endif
}
#endif
/* ================================================================== */
